import os
import sys
from optparse import OptionParser
from StringIO import StringIO
from ase import *
import ase.units as units
#import pylab as plt
from gpaw.utilities.bulk2 import Runner

#structurepacking = {'diamond' }

class Ref:
    def __init__(self, symbol, a, B, e, structure):
        self.symbol = symbol
        self.a = a
        self.B = B
        self.e = e
        self.structure = structure

    def __iter__(self):
        yield self.a
        yield self.B
        yield self.e
        yield self.structure

orderednames = 'FLi C Na MgO Al NaCl Li SiC Si PAl Fe Cu Pt'.split()
orderedbasissets = 'sz szp dz dzp fd'.split()

refs = {
    'C':   Ref('C', 3.575, 431.0, 7.73, 'diamond'),
    'Si':  Ref('Si', 5.459,  87.8, 4.56, 'diamond'),
    'Li':   Ref('Li', 3.438,  13.7, 1.63, 'bcc'),
    'Na':   Ref('Na', 4.200,   7.6, 1.09, 'bcc'),
    'Fe':   Ref('Fe', 2.860, 198.0, 0.00, 'bcc'),
    'Al':   Ref('Al', 4.039,  76.5, 3.42, 'fcc'),
    'Cu':   Ref('Cu', 3.628, 143.9, 99.99, 'fcc'),
    'Pt':   Ref('Pt', 3.999, 234.0, 0.00, 'fcc'),
    'FLi':  Ref('FLi', 4.064,  71.7, 4.13, 'rocksalt'),
    'MgO':  Ref('MgO', 4.259, 147.8, 5.06, 'rocksalt'),
    'NaCl': Ref('NaCl', 5.714,  23.7, 3.13, 'rocksalt'),
    'SiC':  Ref('SiC', 4.381, 211.8, 6.42, 'zincblende'),
    'PAl':  Ref('PAl', 5.513,  82.6, 4.06, 'zincblende')}

def build_parser():
    p = OptionParser(usage='%prog [--action ref|ascii|latex] FILE...')
    p.add_option('--action', default='ref')
    return p

def splitfilename(filename):
    name, tag = filename[:-5].split('-', 1)
    return name, tag


def file2ref(filename):
    configs = read(filename, ':')
    #except IOError:
    #    continue
    #if len(configs) != 5:
    #    continue
    if len(configs) != 5:
        raise ValueError('len(configs) is not 5')
    name, tag = splitfilename(filename)
    #if name in 'Fe Cu Pt'.split():
    #if name == 'Fe':
    #    raise ValueError
    
    runner = Runner(name, configs[2], range(5), tag, out=None)
    runner.run()
    runner.atomize()
    a0 = refs[name].a
    e, v, B, ec, a = runner.summary(a0=a0)
    B *= 1e24 / units.kJ
    #da = (a / a0 - 1) * 100
    #dB = (B / B0 - 1) * 100
    #if ec is None:
    #    ec = ''
    #    dec = '      '
    #else:
    #    dec = (ec / ec0 - 1) * 100
    #    ec = '%10.3f' % ec
    #    dec = '%6.2f' % dec
    ref = Ref(name, a, B, ec, 'unknown')
    return ref


def extract_numbers(filenames):
    keys = ['fd.sz', 'lcao.sz', 'lcao.szp', 'lcao.dz', 'lcao.dzp']
    refdicts = {}
    for key in keys:
        refdicts[key] = {}
    for filename in filenames:
        name, tag = splitfilename(filename)
        try:
            ref = file2ref(filename)
        except ValueError:
            continue
        except IOError:
            continue
        refdicts[tag][name] = ref
    return keys, [refdicts[key] for key in keys]

rename = {'FLi' : 'LiF', 'PAl': 'AlP'}
def namefilter(formula):
    return rename.get(formula, formula)


table_fmts = ['%6.2f', '%6.2f', '%6.0f']

class BarsPrinter:
    def __init__(self):
        self.fmts = ['%f'] * 3
        files = {}
        for key in orderedbasissets:
            files[key] = StringIO()
        self.files = files
        self.prop = None

    def begin(self):
        pass

    def print_header(self, keys, prop):
        self.prop = prop

    def printline(self, formula, numberstrings):
        for key, numberstring in zip(orderedbasissets, numberstrings):
            otherprops = {'a' : 'x', 'e' : 'y', 'B' : 'z'}
            string = '%s %s: %s\n' % (formula.strip(),
                                      otherprops[self.prop], numberstring)
            self.files[key].write(string)
            #print >> self.files[key], string

    def printrule(self):
        pass

    def print_mae(self, q, w, e, r):
        pass

    def end(self):
        pass

    def inbetween(self):
        pass

    def finalize(self):
        for key, buffer in self.files.items():
            lines = buffer.getvalue().split('\n')
            lines.sort()
            outfile = open('barsdata.%s.txt' % key, 'w')
            for line in lines:
                print >> outfile, line

class LatexPrinter:
    def __init__(self):
        self.fmts = table_fmts

    def begin(self):
        print r"""
        % COMPUTER GENERATED TABLE
        \begingroup
        \squeezetable
        \begin{table}[p]
        \begin{ruledtabular}
        """

    def print_header(self, keys, prop):
        latexprops = {'a' : '$a$ (\xc3\x85)',
                      'e' : '$E_c$ (eV)',
                      'B' : '$B$ (GPa)' }
        
        print r"""
        \begin{tabular}{llllll}
        &\multicolumn{5}{c}{%s}\\
        \cline{1-6}
        &%s\\
        \hline\hline
        """ % (latexprops[prop], '&'.join([key.upper() for key in keys]))

    def printrule(self):
        pass

    def printline(self, formula, strings):
        print '&'.join([formula] + strings), r'\\'


    def print_mae(self, mae_id, mae_strings, rmae_id, rmae_strings):
        print r'\hline'
        
        print 'MAE &', '&'.join(mae_strings), r'\\'
        print 'MAE \\% &', '&'.join(rmae_strings), r'\\'

    def end(self):
        print r'\end{tabular}'

    def inbetween(self):
        print r'\bigskip'

    def finalize(self):
        print r"""
        \end{ruledtabular}
        \caption{Lattice constants $a$, cohesive energies $E_c$ and bulk
        moduli for selected solids.  Errors are relative to converged
        grid results.  The default basis sets were used.}
        \label{tab:solids-E0.01}
        \end{table}
        \endgroup
        """

class ASCIIPrinter:
    def __init__(self):
        self.headerlength = 0
        self.fmts = table_fmts

    def begin(self):
        pass

    def print_header(self, keys, prop):
        headers = ''.join(['    %s' % prop] + keys)
        self.headerlength = len(headers)
        print headers
        
        
    def printrule(self):
        print '-' * self.headerlength

    def printline(self, formula, strings):
        print formula, ' '.join(strings),
        print

    def print_mae(self, mae_id, mae_strings, rmae_id, rmae_strings):
        print ' '.join([mae_id] +mae_strings)
        print ' '.join([rmae_id] + rmae_strings)

    def end(self):
        print

    def inbetween(self):
        pass

    def finalize(self):
        pass


def print_table(printer, filenames):
    keys, refdicts = extract_numbers(filenames)
    # preserve ordering
    formulas = [name for name in orderednames if name in refdicts[0]]
    keys.append(keys.pop(0))
    refdicts.append(refdicts.pop(0))

    gridrefs = refdicts[-1]

    props = ['a', 'e', 'B']
    fmts = printer.fmts

    fmt_mae = {'a' : '%6.3f', 'e' : '%6.3f', 'B': '%6.2f'}
    fmt_rmae = {'a' : '%6.3f', 'e' : '%6.3f', 'B': '%6.2f'}

    printer.begin()

    for prop, fmt in zip(props, fmts):
        errsums = [0.0] * len(keys)
        relerrsums = [0.0] * len(keys)
        naverages = [0] * len(keys)
        def getprop(ref):
            return getattr(ref, prop)
        printer.print_header([key.split('.')[1].rjust(7)
                              for key in keys[:-1] + ['.grid']], prop)
        printer.printrule()
        #rule = '-' * len(headers)
        #print rule
        for formula in formulas:
            formulastring = namefilter(formula).rjust(5)
            strings = []
            for i, (key, refs) in enumerate(zip(keys, refdicts)):
                ref = refs[formula]
                gridref = gridrefs[formula]
                value = getprop(ref)
                gridvalue = getprop(gridref)
                if value is None:
                    value = float('nan')
                strings.append(fmt % value)

                if gridvalue is not None:
                    abserr = abs(value - gridvalue)
                    relerr = abserr / gridvalue
                    relerrsums[i] += relerr
                    errsums[i] += abserr
                    naverages[i] += 1#[prop] += 1
            #print
            printer.printline(formulastring, strings)
        printer.printrule()

        mae_id = 'MAE'.rjust(5)
        mae_strings = [fmt_mae[prop] % (errsum / naverage)
                       for errsum, naverage in zip(errsums[:-1], naverages)]
        rmae_id = 'MAE %'.rjust(5)
        rmae_strings = [fmt_mae[prop] % (100 * errsum / naverage)
                        for errsum, naverage in zip(relerrsums[:-1],
                                                    naverages)]
                                                    
        #rmae_strings = ' '.join([] +
                                
        printer.print_mae(mae_id, mae_strings, rmae_id, rmae_strings)
        #printer.print_mae(fmt_mae[prop], errsums, relerrsums, naverages)
        printer.end()
        printer.inbetween()
    printer.finalize()

def latextable(filenames):
    pass

def reftable(filenames):
    # hack to separate grid from fd calculations
    fdfiles = []
    lcaofiles = []
    for filename in filenames:
        name, tag = filename[:-5].split('-', 1) # XXX
        if tag.startswith('fd.sz'):
            fdfiles.append(filename)
        else:
            assert tag.startswith('lcao.'), tag
            lcaofiles.append(filename)
    fdfiles.sort()
    lcaofiles.sort()

    for filename in fdfiles + lcaofiles:#sys.argv[1:]:
        try:
            configs = read(filename, ':')
        except IOError:
            continue
        if len(configs) != 5:
            continue
        name, tag = filename[:-5].split('-', 1)
        runner = Runner(name, configs[2], range(5), tag, out=None)
        runner.run()
        runner.atomize()
        a0, B0, ec0, x = refs[name]
        try:
            #e, v, B, ec, a = runner.summary(a0=3.52)
            e, v, B, ec, a = runner.summary(a0=a0)
        except ValueError:
            print '%-7s %-25s eos ValueError' % (name, tag)
            continue
        B *= 1e24 / units.kJ
        da = (a / a0 - 1) * 100
        dB = (B / B0 - 1) * 100
        if ec is None:
            ec = ''
            dec = '      '
        else:
            dec = (ec / ec0 - 1) * 100
            ec = '%10.3f' % ec
            dec = '%6.2f' % dec

    # Output:
    # Symbol, identifier, a err%, B err%, Ecoh err%, a, B, Ecoh

        print ('%-7s %-25s %6.2f %6.2f %s %10.3f %10.3f %s' %
                   (name, tag, da, dB, dec, a, B, ec))


    if 0:#for name, type, h, k in bulk:
        v, e, B, symbols = bulk[(name, type, h, k)]
        B *= 1e24 / kJ

        ec = -e
        for symbol in symbols:
            if (symbol, type, h) in atom:
                ec += atom[(symbol, type, h)]
            else:
                ec = None
                break

        a0, B0, ec0, x = refs[name]
        if x is rocksalt or x is zincblende:
            v0 = x('X', 'X', a0).get_volume()
        else:
            v0 = x('X', a0).get_volume()
        a = a0 * (v / v0)**(1 / 3.0)
        #plt.plot(volumes, energies, 'o')
        #x = np.linspace(volumes[0], volumes[-1], 50)
        #plt.plot(x, eos.fit0(x**-(1.0 / 3)), '-r')
    #plt.show()

def main():
    parser = build_parser()
    opts, args = parser.parse_args()
    if opts.action == 'ref':
        reftable(args)
    else:
        if opts.action == 'ascii':
            printer = ASCIIPrinter()
        elif opts.action == 'latex':
            printer = LatexPrinter()
        elif opts.action == 'bars':
            printer = BarsPrinter()
        else:
            raise ValueError('Bad action %s' % opts.action)
        print_table(printer, args)    

if __name__ == '__main__':
    main()
